import { samePublicKey, publicKey } from '@metaplex-foundation/umi';
import { ProgramNotRecognizedError, ProgramErrorNotRecognizedError } from './errors.mjs';

function createDefaultProgramRepository(context) {
  const programs = [];
  const has = (identifier, clusterFilter = 'current') => {
    const programs = all(clusterFilter);
    return typeof identifier === 'string' ? programs.some(p => p.name === identifier) : programs.some(p => samePublicKey(p.publicKey, identifier));
  };
  const get = (identifier, clusterFilter = 'current') => {
    const cluster = parseClusterFilter(clusterFilter);
    const programs = all(clusterFilter);
    const program = typeof identifier === 'string' ? programs.find(p => p.name === identifier) : programs.find(p => samePublicKey(p.publicKey, identifier));
    if (!program) {
      throw new ProgramNotRecognizedError(identifier, cluster);
    }
    return program;
  };
  const getPublicKey = (identifier, fallback, clusterFilter) => {
    try {
      return get(identifier, clusterFilter).publicKey;
    } catch (error) {
      if (fallback === undefined) throw error;
      return publicKey(fallback);
    }
  };
  const all = (clusterFilter = 'current') => {
    const cluster = parseClusterFilter(clusterFilter);
    return cluster === '*' ? programs : programs.filter(program => program.isOnCluster(cluster));
  };
  const add = (program, overrides = true) => {
    if (overrides) {
      programs.unshift(program);
    } else {
      programs.push(program);
    }
  };
  const resolveError = (error, transaction) => {
    // Ensure the error as logs.
    if (!Array.isArray(error.logs) || error.logs.length === 0) return null;
    const logs = error.logs.join('\n');

    // Parse the instruction number.
    const instructionRegex = /Error processing Instruction (\d+):/;
    const instruction = error.message.match(instructionRegex)?.[1] ?? null;

    // Parse the error code.
    const errorCodeRegex = /Custom program error: (0x[a-f0-9]+)/i;
    const errorCodeString = logs.match(errorCodeRegex)?.[1] ?? null;
    const errorCode = errorCodeString ? parseInt(errorCodeString, 16) : null;

    // Ensure we could find an instruction number and an error code.
    if (instruction === null || errorCode === null) return null;

    // Get the program ID from the instruction in the transaction.
    const instructionNumber = parseInt(instruction, 10);
    const programIndex = transaction.message.instructions?.[instructionNumber]?.programIndex ?? null;
    const programId = programIndex ? transaction.message.accounts[programIndex] : null;

    // Ensure we were able to find a program ID for the instruction.
    if (!programId) return null;

    // Find a registered program if any.
    let program;
    try {
      program = get(programId);
    } catch (_programNotFoundError) {
      return null;
    }

    // Finally, resolve the error.
    const resolvedError = program.getErrorFromCode(errorCode, error);
    return resolvedError ?? new ProgramErrorNotRecognizedError(program, error);
  };
  const parseClusterFilter = clusterFilter => clusterFilter === 'current' ? context.rpc.getCluster() : clusterFilter;
  return {
    has,
    get,
    getPublicKey,
    all,
    add,
    resolveError
  };
}

export { createDefaultProgramRepository };
//# sourceMappingURL=createDefaultProgramRepository.mjs.map
